/** [World Aggregator] A little website written in Golang, HTML, CSS and JS
 * Copyright (C) 2019 Mina Pecheux

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 * -----------------------------------------------------------------------------
 * server.go - Go server with data fetching, data processing and data serving
 * =============================================================================
 **/
package main

/* Import interesting packages */
import (
	"os"
	"fmt"
  "math"
  "strings"
  "strconv"
  "net/http"
	"path/filepath"
  "html/template"
	"sync"
  "golang.org/x/net/html"
  "github.com/antchfx/htmlquery"
)

/* Define the extraction functions */
type ExtractionFunction func(doc *html.Node) ([][]*html.Node)

func WorldometerExtractor(doc *html.Node) ([][]*html.Node) {
	names := htmlquery.Find(doc, "//tbody/tr/td[2]/a")
	values := htmlquery.Find(doc, "//tbody/tr/td[3]")
	return [][]*html.Node{names, values}
}

func WikipediaExtractor0(doc *html.Node) ([][]*html.Node) {
	names := htmlquery.Find(doc, "//table/tbody/tr/td[1]")
	values := htmlquery.Find(doc, "//table/tbody/tr/td[4]")
	names = names[:len(values)]
	return [][]*html.Node{names, values}
}

func WikipediaExtractor1(doc *html.Node) ([][]*html.Node) {
	names := htmlquery.Find(doc, "//table[contains(@class, \"wikitable\")]/tbody/tr/td[1]")
	values := htmlquery.Find(doc, "//table[contains(@class, \"wikitable\")]/tbody/tr/td[5]")
	names = names[:len(values)]
	return [][]*html.Node{names, values}
}

func WikipediaExtractor2(doc *html.Node) ([][]*html.Node) {
	names := htmlquery.Find(doc, "//table[contains(@class, \"wikitable\")]/tbody/tr/td[1]")
	values := htmlquery.Find(doc, "//table[contains(@class, \"wikitable\")]/tbody/tr/td[2]")
	return [][]*html.Node{names, values}
}

func WikipediaExtractor3(doc *html.Node) ([][]*html.Node) {
	names := htmlquery.Find(doc, "//table[contains(@class, \"wikitable\")]/tbody/tr/td[2]")
	values := htmlquery.Find(doc, "//table[contains(@class, \"wikitable\")]/tbody/tr/td[3]")
	return [][]*html.Node{names, values}
}

/* Define the postprocessing functions */
type PostprocessingFunction func(data float64) (float64)

func identity(data float64) (float64) {
  return data
}

func toLog(data float64) (float64) {
  return math.Log10(data)
}

/* Define the search contexts
* (structure and map of URLs to browse) */
type SearchContext struct {
	Url							string
	Extractor				ExtractionFunction
	Postprocessor		PostprocessingFunction
	Unit						string
}

var searchContexts = map[string]SearchContext{
	"Population": SearchContext{"https://www.worldometers.info/world-population/population-by-country/", WorldometerExtractor, toLog, "Number of inhabitants"},
	"GDP": SearchContext{"https://www.worldometers.info/gdp/gdp-by-country/", WorldometerExtractor, toLog, "$"},
	"CO2 Emissions": SearchContext{"https://en.wikipedia.org/wiki/List_of_countries_by_carbon_dioxide_emissions", WikipediaExtractor0, toLog, "MT/year"},
	"Renewable Energy Production": SearchContext{"https://en.wikipedia.org/wiki/List_of_countries_by_electricity_production_from_renewable_sources", WikipediaExtractor1, toLog, "GWh"},
	"Tertiary Education Attainment": SearchContext{"https://en.wikipedia.org/wiki/List_of_countries_by_tertiary_education_attainment", WikipediaExtractor2, identity, "%"},
	"Percentage of Internet Users": SearchContext{"https://en.wikipedia.org/wiki/List_of_countries_by_number_of_Internet_users", WikipediaExtractor1, identity, "%"},
	"Ecological Footprint": SearchContext{"https://en.wikipedia.org/wiki/List_of_countries_by_ecological_footprint", WikipediaExtractor3, identity, "gha/person"},
	"Natural Gas Exports": SearchContext{"https://en.wikipedia.org/wiki/List_of_countries_by_natural_gas_exports", WikipediaExtractor3, toLog, "CU M"},
	"Oil Exports": SearchContext{"https://en.wikipedia.org/wiki/List_of_countries_by_oil_exports", WikipediaExtractor3, toLog, "bbl/day"},
}

/* Data gathering structs */
type GatheredData struct {
	DatasetRef 							string
	DatasetUnit							string
	DatasetPostprocessing		PostprocessingFunction
	Data										[]CountryData
}

type CountryData struct {
	Country    	string
	Value				float64
	NormValue		float64
}

type CountryDataset struct {
	Data 				[]CountryData
	MinValue		float64
	MaxValue		float64
	Unit				string
}

/* Define the WaitGroup for concurrent processing */
var wg sync.WaitGroup

/* Util function to run concurrently (fetches data and
 * adds it to the result channel) */
func getData(c chan GatheredData, context SearchContext, datasetRef string) {
	defer wg.Done()
	var d []CountryData
	
	doc, err := htmlquery.LoadURL(context.Url)
	if (err != nil) {
		panic(err)
	}
	
	results := context.Extractor(doc)
	names := results[0]
	values := results[1]
	
	for i, name := range names {
		n := strings.TrimSpace(htmlquery.InnerText(name))
		val := htmlquery.InnerText(values[i])
		val = strings.TrimSuffix(val, "\n")
		val = strings.TrimSpace(val)
		val = strings.Split(val, " ")[0]
		val = strings.Replace(val, ",", "", -1)
		val = strings.TrimPrefix(val, "$")
		val = strings.TrimSuffix(val, "%")
		v, _ := strconv.ParseFloat(val, 64)
		d = append(d, CountryData{n, v, 0.0})
	}
	
	c <- GatheredData{datasetRef, context.Unit, context.Postprocessor, d}
}

/* Serving function (with concurrent calling of the util function,
 * data processing and HTML templates serving) */
func serveIndex(w http.ResponseWriter, r *http.Request) {
  countryDatasets := make(map[string]CountryDataset)  
	queue := make(chan GatheredData, len(searchContexts))
	
	// loop over search contexts
	for ref, context := range searchContexts {
		wg.Add(1)
		go getData(queue, context, ref)
	}
	
	wg.Wait()
	close(queue)
	
  // loop over results and reconstruct data
	for elem := range queue {
		var min, max float64
		min = math.MaxFloat64
		max = 0
		m := 0.0
		for _, data := range elem.Data {
			if data.Value < min {
				min = data.Value
			}
			if data.Value > max {
				max = data.Value
			}
	
			v := elem.DatasetPostprocessing(data.Value)
			if v > m {
				m = v
			}
		}
		for i, data := range elem.Data {
			v := elem.DatasetPostprocessing(data.Value)
			elem.Data[i].NormValue = v / m
		}
		countryDatasets[elem.DatasetRef] = CountryDataset{elem.Data, min, max, elem.DatasetUnit}
	}
  
  lp := filepath.Join("templates", "layout.html")
  fp := filepath.Join("templates", filepath.Clean(r.URL.Path))

  // Return a 404 if the template doesn't exist
  info, err := os.Stat(fp)
  if err != nil {
    if os.IsNotExist(err) {
      http.NotFound(w, r)
      return
    }
  }

  // Return a 404 if the request is for a directory
  if info.IsDir() {
    http.NotFound(w, r)
    return
  }

  tmpl, err := template.ParseFiles(lp, fp)
  if err != nil {
    // Log the detailed error
    fmt.Println(err.Error())
    // Return a generic "Internal Server Error" message
    http.Error(w, http.StatusText(500), 500)
    return
  }

  if err := tmpl.ExecuteTemplate(w, "layout", countryDatasets); err != nil {
    fmt.Println(err.Error())
    http.Error(w, http.StatusText(500), 500)
  }
}

/* Main entrypoint */
func main() {
	fs := http.FileServer(http.Dir("static"))
  http.Handle("/static/", http.StripPrefix("/static/", fs))
  http.HandleFunc("/", serveIndex)
  http.ListenAndServe(":8000", nil)
}
